/* inbreed -- inbreeding coefficients and kinships
*
*  argv[1] -- pedigree
*  argv[2] -- specified individuals (e.g., potential breeding population)
*  argv[3] -- kinships of founders
*  argv[4] -- 0 = inbreeding coefficients, 1 = kinships, 2 = mean kinships

What it does on Galaxy
The user specifies a pedigree. This is done with a Galaxy table with one row per individual, containing (1) the individual's name, (2) the name of one of the individual's parents, which must have occurred at the start of a previous line, and (3) the name of the individual's other parent, which occurred at the start of a previous line. For a pedigree founder, each parent name is replaced by "-".

Example:

A - -
B - -
C - -
D - -
E - -
F A B
G A B
Thelma A F
Louise F G

Rows can have more than three columns (such as the individual's sex), but only the first three columns affect this command.

The user also provides a file that specifies a set of names of individuals (specifically the first word on each line (one line per individual); any subsequent information on a line is ignored. Finally the user picks from among the options:

Suppose on the other hand that we select a an alternative "founder" set, {A, F, G}. (We require a founder sets to have a member on any ancestral path from Thelma or Louise.) The above pedigree file is then replaced by

A - -
F - -
G - -
Thelma A F
Louise F G

The user theb also provides a file giving kinship information for each pairs of distinct individuals from the founder set; for the current example, the kinship file is as follows:

A A 0.5
F F 0.5
G G 0.6
A F 0.25
A G 0.25
F G 0.25

since an individual's self-kinship is 0.5, while both parent-child pairs and siblings have kinship 0.25. The advantage is that this capability can be used in cases where the kinships of the founders are not initially known, but instead are computationally predicted, e.g., with the Galaxy "Discover" tool.

Finally, the user selects from among the following options:

1. inbreeding coefficients for each specified individual
2. the kinship for each pair of distinct specified individual
3. the mean kinship for each specified individual, i.e., the average kinship value for that individual and every specified individual.

The command reports the requested values.
*/

#include "lib.h"

#define MAX_PED 10000
#define MAX_INDIVIDUALS 10000
#define MAX_FOUNDERS 1000

struct ped {
	char *name;
	int lparent, rparent, breed;
	float coef;
} P[MAX_PED + MAX_INDIVIDUALS];
int nP;

float kin[MAX_FOUNDERS][MAX_FOUNDERS], tot_prob, tot_kinship[1000];

int left_height[1000], done[1000], debug, debug2, root, kinship, nkinship[1000],
    left_founder;

char buf[500];



float negPow2(int n) {
	float x;

	for (x = 1.0; n > 0; --n)
		x *= 0.5;
	return x;
}

int find(char *s) {
	int i;

	if (same_string(s, "-"))
		return -1;
	for (i = 0; i < nP; ++i)
		if (same_string(s, P[i].name))
			return i;
	fatalf("cannot find %s: %s", s, buf);
}

/* given a path from the left parent of root to a non-founder node, we perform
*  a depth-first search from the rigth parent of root. We backtrack from any
*  node that has already been processed by the left_search. At any node on the
*  left_search path, we record the add the contribution to the root's
*  inbreeding coefficient from the cycle.
*/

void right_climb(int n, int height) {
	int L, LL, lft, rgt;
	float prob;

	if (debug2)
		printf("  right_climb %s %d\n", P[n].name, height);
	L = left_height[n];
	if (L < 0)
		return;
	lft = P[n].lparent;
	rgt = P[n].rparent;
	if (lft == -1 && rgt == -1) {
		LL = left_height[left_founder];
		prob = (negPow2(LL+height-2)*kin[left_founder][n]);
		if (debug && prob > 0.0)
			printf("    %s (%d) and %s (%d)  prob = %4.4f\n",
			  P[left_founder].name, LL, P[n].name, height, prob);
		tot_prob += prob;
		return;
	}
	if (L > 0) {
		prob = negPow2(L+height-1)*(1.0+P[n].coef);
		if (debug)
			printf("    %s at heights %d and %d, prob = %4.4f\n",
			  P[n].name, L, height, prob);
		tot_prob += prob;
		done[n] = 1;
		return;
	}
	if (lft != -1)
		right_climb(lft, height+1);
	if (rgt != -1)
		right_climb(rgt, height+1);
}

/* left_climb performs a depth-first search from the left parent of the current
*  root. Each time the path terminates at a founder, n, we perform a depth-first
*  search from the right parent of the current root. At that point, for each
*  node, say k, on the current path from the root's left parent to node n,
*  left_hight[k] is the length of that path-prefix;
*  left_height[k]  = -1 if we have already identified the cycles with the
*  left half formed by a prefix of the currently active left path.
*  left_height[k] = 0 if node k is not on the current path.
*
*  done[k] = 1 if we have already discovered a cycle with apex k and this
*  left half. We'll set left_path[k] = -1 after the right-search completes.
*/

void left_climb(int n, int height) {
	int lft, rgt, i;

	if (debug2)
		printf("  left_climb %s %d\n", P[n].name, height);
	left_height[n] = height;
	lft = P[n].lparent;
	rgt = P[n].rparent;
	if (lft == -1 && rgt == -1) {
		left_founder = n;
		right_climb(P[root].rparent, 1);
		for (i = 0; i < nP; ++i)
			if (done[i])
				left_height[i] = -1;
	}
	if (lft != -1)
		left_climb(lft, height+1);
	if (rgt != -1)
		left_climb(rgt, height+1);
	left_height[n] = done[n] = 0;
}

void get_coef() {
	int lft, rgt, i;

	if (debug2)
		printf("get_coef %s\n", P[root].name);
	lft = P[root].lparent;
	rgt = P[root].rparent;
	if (lft == -1 || rgt == -1) {
		P[root].coef = 0.0;
		//printf("coef at %s = 0.0\n", P[root].name);
		return;
	}

	for (i = 0; i < nP; ++i)
		left_height[i] = done[i] = 0;
	tot_prob = 0.0;
	left_climb(lft, 1);

	P[root].coef = tot_prob;
	//printf("coef at %s = %6.4f\n", P[root].name, tot_prob);
}

void add_node(char *s, int y, int z) {
	P[nP].name = copy_string(s);
	P[nP].lparent = y;
	P[nP].rparent = z;
	++nP;
}

int main(int argc, char **argv) {
	FILE *fp;
	char x[100], y[100], z[100];
	int orig_nP = 0, i, j, k;
	float kinsh;

	if (argc == 6 && same_string(argv[5], "debug"))
		debug = 1;
	else if (argc == 6 && same_string(argv[5], "debug2"))
		debug = debug2 = 1;
	else if (argc != 5)
		fatal("args : pedigree breeders founder's-kinships select");
	
	kinship = atoi(argv[4]);

	// read the pedigree
	fp = ckopen(argv[1], "r");
	while (fgets(buf, 500, fp)) {
		if (buf[0] == '#' || buf[0] == '\n')
			continue;
		if (sscanf(buf, "%s %s %s", x, y, z) != 3)
			fatalf("needs 3 names: %s", buf);
		add_node(x, find(y), find(z));
	}
	fclose(fp);

	// read the potential breeders
	fp = ckopen(argv[2], "r");
	while (fgets(buf, 500, fp)) {
		if (buf[0] == '#' || buf[0] == '\n')
			continue;
		if (sscanf(buf, "%s", x) != 1)
			fatalf("breeder? %s", buf);
		P[find(x)].breed = 1;
	}
	fclose(fp);

	// read the founders' kinships
	for (i = 0; i < MAX_FOUNDERS; ++i)
		for (j = 0; j < MAX_FOUNDERS; ++j)
			kin[i][j] = (i == j ? 0.5 : 0.0);
	fp = ckopen(argv[3], "r");
	while (fgets(buf, 500, fp)) {
		if (buf[0] == '#' || buf[0] == '\n')
			continue;
		if (sscanf(buf, "%s %s %f", x, y, &kinsh) != 3)
			fatalf("kinship? %s", buf);
		kin[find(x)][find(y)] = kinsh;
		kin[find(y)][find(x)] = kinsh;
	}
	fclose(fp);
//printf("find(\"39\") = %d\n", find("39"));
//printf("find(\"30\") = %d\n", find("30"));

	// for kinships, add faux child for every potential breeding pair
	if (kinship) {
		orig_nP = nP;
		for (i = 0; i < orig_nP; ++i) {
			if (!P[i].breed)
				continue;
			for (j = i; j < orig_nP; ++j) {
				if (!P[j].breed)
					continue;
				sprintf(x, "child_%s_%s", P[i].name, P[j].name);
				add_node(x, i, j);
				P[find(x)].breed = 1;
			}
		}
	}

	// compute the inbreeding coefficient for each node, top-down
	for (root = 0; root < nP; ++root)
{
//debug = debug2 = same_string(P[root].name, "titus");
//debug = same_string(P[root].name, "titus");
		get_coef();
}

	if (kinship == 0) {
		for (i = 0; i < nP; ++i)
			if (P[i].breed)
				printf("inbreeding coefficient of %s = %f\n",
				  P[i].name, P[i].coef);
	}

	if (kinship == 1) {
		for (i = orig_nP; i < nP; ++i)
			printf("kinship %s %s = %f\n", P[P[i].lparent].name,
			  P[P[i].rparent].name, P[i].coef);
	}
	if (kinship == 2) {
		for (i = orig_nP; i < nP; ++i) {
			j = P[i].lparent;
			nkinship[j]++;
			tot_kinship[j] += P[i].coef;
			k = P[i].rparent;
			if (k == j)
				continue;
			nkinship[k]++;
			tot_kinship[k] += P[i].coef;
		}
		for (i = 0; i < nP; ++i)
			if (nkinship[i])
				printf("mean kinship of %s = %6.4f\n",
				 P[i].name, tot_kinship[i]/(float)nkinship[i]);
	}
	return 0;
}
