/* kinship_prep -- prepare the ".ped" and ".map" files (PLINK format) for input to
*  the "admixture" program.
*
*  argv[1] -- a Galaxy SNP table
*  argv[2] -- required number of reads for each individual to use a SNP
*  argv[3] -- required genotype quality for each individual to use a SNP
*  argv[4] -- minimum spacing between SNPs on the same scaffold
*  argv[k] for k > 4 have the form "13:fred", meaning that the 13th and 14th
*    columns (base 0) give the allele counts for the individual or group named
*    "fred".

What it does on Galaxy
The user specifies a SNP table (either gd_snp or gd_genotype format) and a set of individuals. The command runs the KING program (Manichaikul et al., 2010) to look for pairs of distinct individuals in the specified set that have a close family relationship. Putatively related pairs are classified into five categories:

duplicate or MZ twin

1st degree relatives -- siblings (other than identical twins) or parent-child

2nd degree relatives -- e.g., half-siblings, grandparent-grandchild pair, individual-uncle/aunt pair

3rd degree relatives -- e.g., first cousins

unrelated

Reference:
Manichaikul A, Mychaleckyj JC, Rich SS, Daly K, Sale M, Chen WM (2010) Robust relationship inference in genome-wide association studies. Bioinformatics 26: 2867-2873.
*/

#include "lib.h"

// bounds line length for a line of the Galaxy table
#define MOST 50000
struct individual {
	int column;
	char *name;
} I[MOST/8]; // each individual has 4 columns and 4 tab characters
int nI;	// number of individuals
int X[MOST];	// integer values in a row of the SNP table

// bounds the number of SNPs that can be kept
#define MAX_KEEP 10000000
char *S[MAX_KEEP];	// S[i] is a row of 2*nI alleles
int nK;

int main(int argc, char **argv) {
	FILE *fp, *ped, *map, *dat;
	char *p, *z = " \t\n", buf[MOST], trash[MOST], name[100], *s,
	  scaf[100], prev_scaf[100];
	int i, j, m, min_coverage, min_quality, min_space, nsnp, genotype,
	   pos, prev_pos;

	if (argc < 6)
		fatal("args: Galaxy-table min-cov min-qual min-space 13:fred 16:mary ...");
	min_coverage = atoi(argv[2]);
	min_quality = atoi(argv[3]);
	min_space = atoi(argv[4]);

	for (i = 5; i < argc; ++i, ++nI) {
		if (nI >= MOST/8)
			fatal("Too many individuals");
		if (sscanf(argv[i], "%d:%s", &(I[nI].column), name) != 2)
			fatalf("bad arg: %s", argv[i]);
		I[nI].name = copy_string(name);
fprintf(stderr, "%d %s\n", nI, name);
	}

	map = ckopen("kinship.map", "w");

	fp = ckopen(argv[1], "r");
	prev_scaf[0] = '\0';
	prev_pos = 0;
	for (nsnp = 0; fgets(buf, MOST, fp); ) {
		if (buf[0] == '#')
			continue;
		++nsnp;
if (nsnp%100000 == 0) fprintf(stderr, "%d\n", nsnp);
		if (sscanf(buf, "%s %d", scaf, &pos) != 2)
			fatalf("choke: %s", buf);
		if (same_string(scaf, prev_scaf)) {
			if (pos < prev_pos + min_space)
				continue;
		} else {
			strcpy(prev_scaf, scaf);
			prev_pos = -min_space;
		}

		// X[i] = atoi(i-th word base-1)
		strcpy(trash, buf);
		for (i = 1, p = strtok(trash, z); p != NULL;
		     ++i, p = strtok(NULL, z))
			X[i] = atoi(p);
		for (i = 0; i < nI; ++i) {
			m = I[i].column;
			if (min_coverage > 0 && X[m] + X[m+1] < min_coverage)
				break;
			if (min_quality > 0 && X[m+3] < min_quality)
				break;
		}
		if (i < nI)
			continue;
		prev_pos = pos;
		
		if (nK >= MAX_KEEP)
			fatal("Too many SNPs");
		fprintf(map, "1 snp%d %d\n", nsnp, nsnp+1);
		s = S[nK++] = ckalloc(2*nI*sizeof(char));
		for (i = j = 0; i < nI; ++i, j += 2) {
			genotype = X[I[i].column+2];
			if (genotype == 2)
				s[j] = s[j+1] = '1';
			else if (genotype == 0)
				s[j] = s[j+1] = '2';
			else if (genotype == 1) {
				s[j] = '1';
				s[j+1] = '2';
			} else	// undefined genotype
				s[j] = s[j+1] = '0';
		}
	}

	fclose(map);

	ped = ckopen("kinship.ped", "w");
	for (i = 0; i < nI; ++i) {
		fprintf(ped, "1 %s 0 0 1", I[i].name);
		for (j = 0; j < nK; ++j)
			fprintf(ped, " %c %c", S[j][2*i], S[j][2*i+1]);
		putc('\n', ped);
	}

	dat = ckopen("kinship.dat", "w");
	for (j = 0; j < nK; ++j)
		fprintf(dat, "M snp%d\n", j);

	printf("Using %d of %d SNPs\n", nK, nsnp);
	fclose(ped);

	return 0;
}
