/* mito_draw -- prepare data for drawing a picture of mitogenome data
*
*  argv[1] -- gd_genotype table for the mitogenome SNVs
*  argv[2] -- coverage table in gd_genotype-like format; for each mitochodrial
*  position (in order) there is a line like:

chrM	1234	A	G	19	0	11 ...

*  argv[3] -- the annotation file, like

AJ428849	0	70	tRNA	+	tRNA-Phe
AJ428849	70	1030	rRNA	+	12S
...
AJ428849	13554	14078	CDS	-	NADH1
...
AJ428849  	15484	16910	D-loop	+	D-loop

*  argv[4] -- the minimum coverage. Intervals of lower coverage are ignored.
*  argv[5], argv[6], ... type:column:name pairs like "i:7:sam" for i = 0, 1, 2.
*    0 means the set of individuals to be drawn
*    1 for the SNP table and
*    2 for the coverage table. 
*
* What it does on Galaxy
The user supplies the following>
1. A optional  file in gd_genotype or gd_snp format giving the mitochondrial SNPs.
2. An optional gd_genotype file gives the sequence coverage for each individual at each mitochondrial position.
3. A file of annotation for the reference mitochondrial sequence.
4. A set of individuals specified sith the "Specify individuals" tool.
5. The minimum depth of sequence coverage. Positions where an individual has less coverage are ignoried.

The program draws a picture indicating the locations of SNPs and the inadequately covered interval.

*/

#include "lib.h"
#include "mito_lib.h"

// print the i-th individual's SNPs that are in adequately covered intervals
void visible_snps(int i) {
	struct snp *s;
	struct covered *c;
	int pos;

	c = I[i].cov;
	for (s = I[i].snp; (pos = s->pos) < ref_len; s = s->next) {
		while (c->e <= pos)
			if ((c = c->next) == NULL)
				fatal("cannot happen");
		if (pos >= c->b)
			printf(" %d", s->pos);
	}
}


// read gene annotation, change "CDS" to "gene", print for the drawing tool,
// print lines showing the genome name and length
void get_genes(char *filename) {
	FILE *fp = ckopen(filename, "r");
	char buf[500], ref[100], type[100], name[100], *t;
	int b, e;

	while (fgets(buf, 500, fp)) {
		if (sscanf(buf, "%s %d %d %s %*s %s",
		    ref, &b, &e, type, name) != 5)
			fatalf("gag: %s", buf);
		t = (same_string(type, "CDS") ? "gene" : type);
		// print the Genome Annotation line
		printf("@GA=%d:%d:%s:%s\n", b, e, name, t);
		ref_len = MAX(ref_len, e+1);
	}
	fclose(fp);
	printf("@GL=%d\n", e);
	printf("@GN=%s\n", ref);
}

int main(int argc, char **argv) {
	int i, old_e;
	struct covered *c;
	// int last_e, refcol, seen_missing = 0, seen_special = 0;

	if (argc < 5)
		fatal("args: snps coverage genes min_cov 0:9:sam 1:13:judy ... ");

	// store sample names and col positions (argv[5], argv[6], ...)
	for (i = 5; i < argc; ++i)
		insert_indiv(argv[i]);

	min_cov = atoi(argv[4]); // must precede the call to get_covered

	// record the variants
	get_snps(argv[1]);

	// record the regions with adequate coverage
	get_covered(argv[2]);

	// read annotation and annotate in the file for drawing;
	get_genes(argv[3]);
	sentinels();	// put end-markers on lists ofSNPs and covered intervals


	// report the information for each sample
	for (i = 0; i < nI; ++i) {
		printf("%s", I[i].name);
		visible_snps(i);
		for (old_e = 0, c = I[i].cov; c; c = c->next) {
			if (old_e < c->b)
				printf(" missing=%d:%d", old_e, c->b);
			old_e = c->e;
		}
		putchar('\n');
	}

	// record color information
	printf("@CL=rRNA:#EF8A62\n@CL=tRNA:#31A354\n@CL=gene:#B2182B\n");
	printf("@CL=missing:#67A9CF:L\n");
#ifdef NEVER
	if (seen_indel)
		printf("@CL=indel:#B2182B\n");
#endif

	return 0;
}
