// mito_lib.c -- procedures to read mitochondrial genotypes and coverage.
// See mito_lib.h for further explanations.

#include "lib.h"
#include "mito_lib.h"

// maximum number of columns in a (gd_genotype) SNP or coverage table
#define MAX_COL 20000
static int X[MAX_COL];
static char *z = " \t\n";

// add covered-interval [b,e] for the i-th individual
static void add_cov(int i, int b, int e) {
	struct covered *c;

	c = ckalloc(sizeof(*c));
	c->b = b;
	c->e = e;
	c->next = NULL;
	if (I[i].cov == NULL)
		I[i].cov = c;
	else
		I[i].last_cov->next = c;
	I[i].last_cov = c;
}

// Store the adequately covered intervals for each individual; update ref_len.
void get_covered(char *filename) {
	FILE *fp = ckopen(filename, "r");
	int col, i, j, x, pos, old_pos;
	char buf[500], *p;

	for (i = 0; i < nI; ++i)
		I[i].start_cov = -1;
	pos = -1;
	while (fgets(buf, 500, fp)) {
		if (buf[0] == '#')
			continue;
		// X[i] = atoi(i-th word, base 1)
		for (i = 1, p = strtok(buf, z); p != NULL;
		     ++i, p = strtok(NULL, z))
			X[i] = atoi(p);
		old_pos = pos;
		pos = X[2];
		if (pos < 0 || pos != old_pos+1)
			fatalf("coverage positions not consecutive at %d",
			  pos);
		for (i = 0; i < nI; ++i)
			if ((col = I[i].cov_col) > 0) {
				x = X[col];
				if (x >= min_cov) {
					if (I[i].start_cov < 0)
						I[i].start_cov = pos;
				} else if (I[i].start_cov >= 0) {
					add_cov(i, I[i].start_cov, pos);
					I[i].start_cov = -1;
				}
			}
	}
	fclose(fp);
	ref_len = MAX(ref_len, pos+1);
	for (i = 0; i < nI; ++i) {
		if (pos == -1) // no coverage file
			add_cov(i, 0, ref_len);
		else if ((j = I[i].start_cov) > -1) // unfinished interval
			add_cov(i, j, ref_len);
	}
}

// add a SNP at pos for the i-th individual
static void add_snp(int i, int pos) {
	struct snp *s;

	s = ckalloc(sizeof(*s));
	s->pos = pos;
	s->next = NULL;
	if (I[i].snp == NULL)
		I[i].snp = s;
	else
		I[i].last_snp->next = s;
	I[i].last_snp = s;
}
	

// store SNPs from the genotypes file; update ref_len
void get_snps(char *filename) {
	FILE *fp = ckopen(filename, "r");
	int i, x, pos, old_pos;
	char buf[500], *p;

	pos = -1;
	while (fgets(buf, 500, fp)) {
		if (buf[0] == '#')
			continue;
		// X[i] = atoi(i-th word, base 1)
		for (i = 1, p = strtok(buf, z); p != NULL;
		     ++i, p = strtok(NULL, z))
			X[i] = atoi(p);
		old_pos = pos;
		pos = X[2];
		if (pos < 0)
			fatalf("negative SNP position: %s", pos);
		if (pos <= old_pos)
			fatalf("SNP out of order at position %d", pos);
		for (i = 0; i < nI; ++i) {
			x = X[I[i].snp_col];
			if (x == 1 || x == 0)
				add_snp(i, pos);
		}
	}
	fclose(fp);
	ref_len = MAX(ref_len, pos+1);
}

// record the name of an individual to be processed, or save the individual's
// columns numbers in the files giving SNPs or coverage information
void insert_indiv(char *argv) {
	char name[100];
	int i, type, col;
	static int seen_positive = 0;

	if (sscanf(argv, "%d:%d:%s", &type, &col, name) != 3)
		fatalf("ouch: %s", argv);
	if (type < 0 || type > 2 || col <= 0)
		fatalf("bad spec: %s", argv);

	// set i to the name's index in the table; insert new names
	for (i = 0; i < nI && !same_string(name, I[i].name); ++i)
		;
	if (type == 0) {
		if (seen_positive)
			fatalf("%s is preceded by a non-zero type", argv);
		if (i == nI)
			I[nI++].name = copy_string(name);
		else
			fatalf("%s is specified twice");
		return;
	}
	if (i == nI)
		return;
	seen_positive = 1;
 	if (type == 1) {
		// which column in the SNP table has the genotype
		if (I[i].snp_col > 0)
			fatalf("SNP has two columns for %s", I[i].name);
		I[i].snp_col = col+2;
	} else if (type == 2) {
		seen_positive = 1;
		// which column in the coverage table has this individual's coverage?
		if (I[i].cov_col > 0)
			fatalf("coverage has two columns for %s", I[i].name);
		I[i].cov_col = col+2;
	} else
		fatal("cannot happen");
}

void sentinels() {
	int i;

	// add a bogus elements to signify the end of each list
	for (i = 0; i < nI; ++i) {
		add_snp(i, ref_len);
		add_cov(i, ref_len, ref_len);
	}
}
