/* mito_pi -- compute the diversity measure pi for mitochondrial genomes
*  [SHOULD I OPTIONALLY INCLUDE INDELS?]
*
*  argv[1] -- SNP table for the mitogenome
*  argv[2] -- coverage table in gd_genotype-like format;  for each mitochodrial
*  position (in order) there is a line like:

chrM	1234	A	G	19	0	11 ...

*  argv[3] -- the minimum coverage; intervals of lower coverage are ignored.
*  argv[4] -- lower bound for shared well-covered bp
*  argv[5], argv[6], ... type:column:name pairs like "i:7:sam" for i = 0, 1, 2.
*     0 indicates the set of individuals to be processed;
*     1 for the SNP table;
*     2 for the coverage table.
*  
*  Also, if the last argument is "debug", then much output is sent to stderr.
*
*  What it does on Galaxy
The user supplies the following:
1. A file in gd_genotype or gd_snp format giving the mitochondrial SNPs.
2. An optional gd_genotype file gives the sequence coverage for each individual at each mitochondrial position.
3. A set of individuals specified with the "Specify individuals" tool.
4. The minimum depth of sequence coverage. Positions where an individual has less coverage are ignored.
5. The number of adequately covered positions that must be shared by two individuals before their diversity is included in the reported average.

For each pair of individual (with adequate shared coverage), the program divides the number of nucleotide difference between the individuals in those intervals by the intervals' total length. Those ratios are averaged over the relevant pairs of individuals.
*/

#include "lib.h"
#include "mito_lib.h"

struct info {
	char *txt;
	struct info *next;
} *first_info = NULL, *last_info;
static int B, E;

// Find the intersection [B,E] of the current two adequately covered intervals.
// Return the interval's length
int olap(struct covered *c, struct covered *d) {
	B = MAX(c->b, d->b);
	E = MIN(c->e, d->e);
	return MAX(E-B, 0);
}

// for a pair of samples, determine how much of the reference is in the
// adequately covered segments for both, and count the number of SNPs in those
// shared regions where the samples differ
float pair(int i, int j, int required) {
	char buf[500];
	int covered, diffs, min_pos;
	struct snp *s, *t;
	struct covered *c, *d;
	struct info *new;
	float pi;

	s = I[i].snp;
	t = I[j].snp;
	c = I[i].cov;
	d = I[j].cov;
	covered = olap(c, d);
	diffs = 0;
	while (B < ref_len) {
		if ((min_pos = MIN(s->pos, t->pos)) < E) {
			if (min_pos >= B)
				if (s->pos != t->pos)
					++diffs;
			if (s->pos == min_pos)
				s = s->next;
			if (t->pos == min_pos)
				t = t->next;
		} else {
			if (c->e == E)
				c = c->next;
			if (d->e == E)
				d = d->next;
			covered += olap(c, d);
		}
	}
	pi = (covered >= required ?  (float)diffs/(float)covered : -1.0);
	sprintf(buf, "%s,%s : %5d bp covered, %4d differences,  pi = %4.4f",
		  I[i].name, I[j].name, covered, diffs, pi);
	new = ckalloc(sizeof(*new));
	new->txt = copy_string(buf);
	new->next = NULL;
	if (first_info == NULL)
		first_info = new;
	else
		last_info->next = new;
	last_info = new;

	return pi;
}

int main(int argc, char **argv) {
	int i, j, good_pairs, bad_pairs, required;
	float tot, pr;

	if (argc > 5 && same_string(argv[argc-1], "debug")) {
		--argc;
		debug = 1;
	}

	if (argc < 5)
		fatal("args: snps coverage min_cov 0:9:sam 1:13:judy ... ");
	// store sample names and start positions (argv[4], argv[5], ...)
	
	for (i = 5; i < argc; ++i)
		insert_indiv(argv[i]);

	if (nI < 2)
		fatal("needs at least two individuals");

	// record the SNPs, covered intervals; set ref_len
	min_cov = atoi(argv[3]); // must precede call to get_covered()
	required = atoi(argv[4]);
	get_snps(argv[1]);
	get_covered(argv[2]);
	sentinels(); // put end-markers on lists of SNPs and covered intervals

	if (debug) {
		struct covered *m;
		struct snp *s;

		fprintf(stderr, "ref_len = %d\n", ref_len);
		for (i = 0; i < nI; ++i) {
			fprintf(stderr, "%s: %d %d\n",
			  I[i].name, I[i].snp_col, I[i].cov_col);
			for (s = I[i].snp; s; s = s->next)
				fprintf(stderr, " %d", s->pos);
			fputc('\n', stderr);
			for (m = I[i].cov; m; m = m->next)
				fprintf(stderr, " [%d,%d]", m->b, m->e);
			fputc('\n', stderr);
		}
	}

	// record the total rate of diversity, over all pairs of individuals
	// having overlapping well-covered intervals
	good_pairs = bad_pairs = 0;
	for (i = 0, tot = 0.0; i < nI; ++i) {
		for (j = i+1; j < nI; ++j) {
			pr = pair(i, j, required);
			if (pr >= 0.0) {
				++good_pairs;
				tot += pr;
			} else
				++bad_pairs;
		}
	}
	if (good_pairs > 0)
		printf("pi = %5.5f\n", tot/(float)good_pairs);
	if (bad_pairs > 0)
		printf("%d of %d pairs had insufficient shared regions\n",
		  bad_pairs, bad_pairs + good_pairs);
	printf("\n\n");
	while (first_info) {
		printf("%s\n", first_info->txt);
		first_info = first_info->next;
	}

	return 0;
}
