/* offspring_heterozygosity -- compute the "offspring expected heterozygosity"
*  between every individual in the first specified set and every individual in
*  the second specified set.
*
*  argv[1] -- a Galaxy SNP table
*  argv[k] for k > 1 have the form "13:0:fred" or "13:1:fred" meaning that the
*  13th and 14th columns (base 0) give the allele counts for the individual
*  named "fred", and that fred is in set 0 or set 1.

What it does on Galaxy
For each pair of individuals, one from each specified set, the program computes the expected heterozygosity of any offspring of the pair, i.e., the probability that the offspring has distinct nucleotides at a randomly chosen autosomal SNP. In other words, we add the following numbers for each autosomal SNP where both genotypes are defined, then divide by the number of those SNPs:

0 if the individuals are homozygous for the same nucleotide
1 if the individuals are homozygous for different nucleotides
1/2 otherwise (i.e., if one or both individuals are heterozygous) 

A SNP is ignorned if one or both individuals have an undefined genotype (designated as -1).
*/

#include "lib.h"

// bounds line length for a line of the Galaxy table
#define MOST 50000
struct individual {
	int column;
	char *name;
} I[MOST], J[MOST];

int nI, nJ,	// number of individuals
    X[MOST],	// integer values in a row of the SNP table
    **N,	// N[x][y] = twice the above count for individuals x and y
    **S;	// S[x][y] = the number of SNPs with defined genotypes

int main(int argc, char **argv) {
	FILE *fp;
	char *p, *z = " \t\n", buf[MOST], name[100];
	int i, j, gi, gj, k, first, col, group;

	if (argc < 3)
		fatal("args: Galaxy-table 13:0:fred 17:1:mary ...");

	// record the two sets of specified individuals
	for (i = 2; i < argc; ++i) {
		if (sscanf(argv[i], "%d:%d:%s", &col, &group, name) != 3)
			fatalf("bad arg: %s", argv[i]);
		if (group == 0) {
			if (nI >= MOST)
				fatal("Too many individuals");
			I[nI].column = col;
			I[nI].name = copy_string(name);
			++nI;
		} else if (group == 1) {
			if (nJ >= MOST)
				fatal("Too many individuals");
			J[nJ].column = col;
			J[nJ].name = copy_string(name);
			++nJ;
		} else
			fatalf("group must be 0 or 1: %s", argv[1]);
	}
	if (nI == 0 || nJ == 0)
		fatal("needs two non-empty sets of individuals");

	// allocate and initialize arrays N and S
	N = ckalloc(nI*sizeof(int *));
	for (i = 0; i < nI; ++i)
		N[i] = ckalloc(nJ*sizeof(int));
	S = ckalloc(nI*sizeof(int *));
	for (i = 0; i < nI; ++i)
		S[i] = ckalloc(nJ*sizeof(int));
	for (i = 0; i < nI; ++i)
		for (j = 0; j < nJ; ++j)
			N[i][j] = S[i][j] = 0;

	// loop over the SNPs
	fp = ckopen(argv[1], "r");
	first = 1;
	while (fgets(buf, MOST, fp)) {
		if (buf[0] == '#')
			continue;
		// X[i] = atoi(i-th word base-1)
		for (i = 1, p = strtok(buf, z); p != NULL;
		     ++i, p = strtok(NULL, z))
			X[i] = atoi(p);
		if (first) {
			for (j = 0; j < nI; ++j)
				if (I[j].column+2 >= i)
					fatalf("only %d SNP-table columns",
					  i-1);
			first = 0;
		}
		for (i = 0; i < nI; ++i) {
			for (j = 0; j < nJ; ++j) {
				gi = X[I[i].column+2];
				gj = X[J[j].column+2];
				if (gi == -1 || gj == -1)
					// undefined genotype: ignore SNP
					continue;
				S[i][j]++;
				if (gi == 1 || gj == 1) 
					// one or both are heterzygous
					k = 1;
				else
					k = (gi == gj ? 0 : 2);
				N[i][j] += k;
			}
		}
	}

	// print the output
	for (i = 0; i < nI; ++i)
		for (j = 0; j < nJ; ++j)
			printf("%s\t%s\t%f\n", I[i].name, J[j].name,
			  (float)N[i][j]/((float)S[i][j]*2.0));

	return 0;
}
