/* offspring_heterozygosity2 -- compute the "offspring expected heterozygosity"
*  between every individual in the first specified set and every individual in
*  the second specified set, given genotyes for the pedigree's founders.
*
*  argv[1] -- a Galaxy SNP table
*  argv[2] -- a pedigree, where the SNP table is for the founders
*  argv[k] for k > 2 have the form "founder:13:fred", "0:fred" or "1:fred".
*    The first form means fred is a founder and in the SNP table with genotype
*    at column 13+2 = 15 (base-1). The other forms mean that fred is in the
*    pedigree and one of the two specified sets of individuals (set 0 or set 1).

What it does on Galaxy
The user provides a Galaxy SNP table (gd_snp or gd_genotype format) that includes the founders of a pedigree, as well as two sets of individuals. The pedigree is specified by a text file with one row per individual, containing (1) the individual's name, (2) the name of one of the individual's parents, which must have occurred at the start of a previous line, and (3) the name of the individual's other parent, which occurred at the start of a previous line. For a pedigree founder, both parent names are replaced by "-". The founders are specified by a table in gd_indivs format, e.g., as produced by "Specify individuals" tool. Every founder must have genotypes supplied in the SNP table, and both parents need to be given as "-" in the pedigree. Conversely, every pedigree individual whose parents are "-" must be named as a founder.

The user also provides two files that specify a set of names of individuals. The first word on each line names an individual (one line per individual); any subsequent information on a line is ignored. The name of each individual must appear at the start of a line in the pedigree.

For each pair of individuals, one from each specified set, the program computes the expected heterozygosity of any offspring of the pair, i.e., the probability that the offspring has distinct nucleotides at a randomly chosen autosomal SNP. A SNP is ignorned if one or both potential parents have an ancestor with undefined genotype (designated as -1 in the SNP table).
*/

#include "lib.h"

// bounds line length for a line of the Galaxy table and other items
#define MOST 50000
struct founder {
	int column, ped_loc;
	char *name;
} F[MOST];

struct individual {
	int ped_loc;
	char *name;
} I[MOST], J[MOST]; // the two specified sets

struct ped {
	char *name;
	int lparent, rparent,	// pedigree locations of male and female parents
	    initialize;		// genotypes provide by the SNP table
	float R, V, T; // probabilities of Reference, Variant, heTerozygous
} P[MOST];

char buf[MOST];

int nF, nI, nJ,	// number of fouders and individuals in the two specified sets
    nP,		// number of individuals in the pedigree
    X[MOST],	// integer values in a row of the SNP table
    **N;	// N[x][y] = the number of SNPs with defined genotypes
		// where x is the first specified set and y is in the second

float **S;	// S[x][y] = total T's of the N[x][y] individuals

// find the pedigree line number for a named individual
int find(char *s) {
	int i;

	if (same_string(s, "-"))	// founder
		return -1;
	for (i = 0; i < nP; ++i)
		if (same_string(s, P[i].name))
			return i;
	fatalf("cannot find %s: %s", s, buf);
}

// insert information about a specified individual in the appropriate table
int add_individual(struct individual *X, int nX, char *name, int p) {
	if (nX >= MOST)
		fatal("Too many individuals");
	X[nX].name = copy_string(name);
	X[nX].ped_loc = p;
	return nX + 1;
}

// compute the probability of being homozygous for the Reference allele
float get_R(float Rm, float Tm, float Rf, float Tf) {
	return Rm*Rf + Rm*Tf/2.0 + Tm*Rf/2.0 + Tm*Tf/4.0;
}

// compute the probability of being homozygous for the Variant allele
float get_V(float Vm, float Tm, float Vf, float Tf) {
	return Vm*Vf + Vm*Tf/2.0 + Tm*Vf/2.0 + Tm*Tf/4.0;
}

int main(int argc, char **argv) {
	FILE *fp;
	char *s, *z = " \t\n", buf[MOST], name[100], u[100], v[100], w[100];
	int i, j, m, f, gi, which_snp, col, group, p, debug = 0, bad;
	float Rf, Vf, Tf, Rm, Vm, Tm, Rx, Vx;

	if (argc < 4)
		fatal("args: Galaxy-table pedigree founder:13:fred 0:mary ...");
	if (same_string(argv[argc-1], "debug")) {
		debug = 1;
		--argc;
	}

	// read the pedigree
	fp = ckopen(argv[2], "r");
	while (fgets(buf, MOST, fp)) {
		if (buf[0] == '#' || buf[0] == '\n')
			continue;
		if (sscanf(buf, "%s %s %s", u, v, w) != 3)
			fatalf("needs 3 names: %s", buf);
		P[nP].name = copy_string(u);
		P[nP].lparent = i = find(v);
		P[nP].rparent = j = find(w);
		if ((i < 0 && j >= 0) || (i >=0 && j < 0))
			fatal("if one parent is \"-\", then both must be \"-\"");
		++nP;
	}
	fclose(fp);

	// record the founders and the two sets of specified individuals
	for (i = 3; i < argc; ++i) {
		if (sscanf(argv[i], "founder:%d:%s", &col,name) == 2) {
			p = find(name);
			//  SNP table must
			if (P[p].lparent != -1)
				fatalf("founder %s has parents in the pedigree",
				   name);
			P[p].initialize = 1;
			if (nF >= MOST)
				fatal("Too many individuals");
			F[nF].column = col;
			F[nF].name = copy_string(name);
			F[nF].ped_loc = p;
			++nF;
		} else if (sscanf(argv[i], "%d:%s", &group, name) == 2) {
			p = find(name);
			if (group == 0)
				nI = add_individual(I, nI, name, p);
			else if (group == 1)
				nJ = add_individual(J, nJ, name, p);
			else
				fatal("the group number must be 0 or 1");
		}
	}
	if (nF == 0 || nI == 0 || nJ == 0)
		fatal("needs founders and two non-empty sets of individuals");

	// make sure that every pedigree founder is in the SNP table
	for (p = bad = 0; p < nP; ++p)
		if (P[p].lparent == -1 && !P[p].initialize) {
			bad = 1;
			fprintf(stderr,
			  "%s has no parents in the pedigree,", P[p].name);
			fprintf(stderr, " but isn't in the SNP table\n");
		}
	if (bad)
		return 1;

	// allocate and initialize arrays N and S
	N = ckalloc(nI*sizeof(int *));
	for (i = 0; i < nI; ++i)
		N[i] = ckalloc(nJ*sizeof(int));
	S = ckalloc(nI*sizeof(float *));
	for (i = 0; i < nI; ++i)
		S[i] = ckalloc(nJ*sizeof(float));
	for (i = 0; i < nI; ++i)
		for (j = 0; j < nJ; ++j) {
			N[i][j] = 0;
			S[i][j] = 0.0;
		}

	// loop over the SNPs
	fp = ckopen(argv[1], "r");
	for (which_snp = 0; fgets(buf, MOST, fp); ) {
		if (buf[0] == '#')
			continue;
		++which_snp;
		if (debug)
			printf("SNP %d: %s", which_snp, buf);
		// X[i] = atoi(i-th word base-1)
		for (col = 1, s = strtok(buf, z); s != NULL;
		     ++col, s = strtok(NULL, z))
			X[col] = atoi(s);
		if (which_snp == 1)
			for (j = 0; j < nI; ++j)
				if (F[j].column+2 >= col)
					fatalf("only %d SNP-table columns",
					  col-1);
		// initialize
		for (i = 0; i < nP; ++i)
			P[i].R = P[i].V = P[i].T = -3.0;	// not set
		// set R, V and T for each founder
		for (i = 0; i < nF; ++i) {
			gi = X[F[i].column+2];
			p = F[i].ped_loc;
			if (gi == -1) {
				// mark as missing genotype
				P[p].R = P[p].V = P[p].T = -1.0;
				continue;
			}
			P[p].R = P[p].V = P[p].T = 0.0;
			if (gi == 2)
				P[p].R = 1.0;
			else if (gi == 0)
				P[p].V = 1.0;
			else
				P[p].T = 1.0;
		}

		// propagate values down the pedigree
		for (i = 0; i < nP; ++i) {
			m = P[i].lparent; // m for male
			f = P[i].rparent;  // f for female
			if (m == -1)
				continue;
			Rm = P[m].R;
			Vm = P[m].V;
			Tm = P[m].T;
			Rf = P[f].R;
			Vf = P[f].V;
			Tf = P[f].T;
			if (Rm < -2.0 || Rf < -2.0) 
				// a parent hasn't been processed
				fatalf("bad pedigree line %s %s %s",
				  P[i].name, P[m].name, P[f].name);
			if (Rm < 0.0 || Rf < 0.0) {
				// undefined genotype
				P[i].R = P[i].V = P[i].T = -1.0;
				continue;
			}
			P[i].R = get_R(Rm, Tm, Rf, Tf);
			P[i].V = get_V(Vm, Tm, Vf, Tf);
			P[i].T = 1.0 - P[i].R - P[i].V;
		}
		if (debug) {
			for (i = 0; i < nP; ++i)
				printf("%s\t%s\t%s\t%1.4f\t%1.4f\t%1.4f\n",
				  P[i].name,
				  ((m = P[i].lparent) >= 0 ? P[m].name : "-"),
				  ((f = P[i].rparent) >= 0 ? P[f].name : "-"),
				  P[i].R, P[i].V, P[i].T);
			putchar('\n');
		}

		// consider potential offspring;
		for (i = 0; i < nI; ++i) {
			m = I[i].ped_loc;
			Rm = P[m].R;
			if (Rm < 0.0) // skip undefined genotype
				continue;
			Vm = P[m].V;
			Tm = P[m].T;
			for (j = 0; j < nJ; ++j) {
				f = J[j].ped_loc;
				Rf = P[f].R;
				if (Rf < 0.0)
					continue;
				Vf = P[f].V;
				Tf = P[f].T;
				++N[i][j];
				Rx = get_R(Rm, Tm, Rf, Tf);
				Vx = get_V(Vm, Tm, Vf, Tf);
				S[i][j] += (1.0 - Rx - Vx);
			}
		}
	}

	// print the output
	for (i = 0; i < nI; ++i)
		for (j = 0; j < nJ; ++j)
			printf("%s\t%s\t%f\n", I[i].name, J[j].name,
			  S[i][j]/(float)N[i][j]);

	return 0;
}
