/* pw_homozygosity -- compute the "pairwise homozygosity" between every pair
*  of specified individuals (including an individual and itself).
*
*  argv[1] -- a Galaxy SNP table
*  argv[k] for k > 2 have the form "13:fred", meaning that the 13th and 14th
*    columns (base 0) give the allele counts for the individual or group named
*    "fred".

What it does on Galaxy
For each pair of specifiedindividuals (including an individual and its), compute the "pairwise homozygosity", i.e., the probability that a randomly chosen allele from each individual at a randomly chosen autosomal SNP have the same nucleotide. In other words, we add the following numbers for each autosomal SNP where both genotypes are defined and divide by the number of those SNPs:

1 if the individuals are homozygous for the same nucleotide
0 if the individuals are homozygous for different nucleotides
1/2 otherwise 

*/

#include "lib.h"

// bounds line length for a line of the Galaxy table
#define MOST 50000
struct individual {
	int column;
	char *name;
} I[MOST/2]; // each individual has at least 1 column and 1 tab character

int nI,		// number of individuals
    X[MOST],	// integer values in a row of the SNP table
    *N,		// N[loc(x,y)] = twice the above count for individuals x and y
    *S;		// S[loc(x,y)] = the number of SNPs with defined genotypes

// Compute location for the total for individuals i and j.
// store only above-diagonal entries
int loc(int i, int j) {
	if (i > j)
		fatal("storing only above-diagonal entries");
	//return i*nI + j - (i*(i-1))/2;
	return i*nI + j;
}

int main(int argc, char **argv) {
	FILE *fp;
	char *p, *z = " \t\n", buf[MOST], name[100];
	int i, j, gi, gj, k, alloc_size, first;

	if (argc < 3)
		fatal("args: Galaxy-table 13:fred 17:mary ...");

	for (i = 2; i < argc; ++i, ++nI) {
		if (nI >= MOST/2)
			fatal("Too many individuals");
		if (sscanf(argv[i], "%d:%s", &(I[nI].column), name) != 2)
			fatalf("bad arg: %s", argv[i]);
		I[nI].name = copy_string(name);
	}

	//alloc_size = nI*(nI+1)/2;
	alloc_size = nI*nI;
	N = ckalloc(alloc_size*sizeof(int));
	S = ckalloc(alloc_size*sizeof(int));
	for (i = 0; i < alloc_size; ++i)
		N[i] = S[i] = 0;

	fp = ckopen(argv[1], "r");
	first = 1;
	while (fgets(buf, MOST, fp)) {
		if (buf[0] == '#')
			continue;
		// X[i] = atoi(i-th word base-1)
		for (i = 1, p = strtok(buf, z); p != NULL;
		     ++i, p = strtok(NULL, z))
			X[i] = atoi(p);
		if (first) {
			for (j = 0; j < nI; ++j)
				if (I[j].column+2 >= i)
					fatalf("only %d SNP-table columns",
					  i-1);
			first = 0;
		}
		for (i = 0; i < nI; ++i) {
			for (j = i; j < nI; ++j) {
				gi = X[I[i].column+2];
				gj = X[I[j].column+2];
				if (gi == -1 || gj == -1)
					continue;
				++S[loc(i,j)];
				if (gi == 1 || gj == 1)
					k = 1;
				else
					k = (gi == gj ? 2 : 0);
				N[loc(i,j)] += k;
			}
		}
	}

	for (i = 0; i < nI; ++i)
		for (j = i; j < nI; ++j)
			printf("%s %s %f\n", I[i].name, I[j].name,
			  (float)N[loc(i,j)]/((float)S[loc(i,j)]*2.0));

	return 0;
}
